package org.xiss.index;

import java.util.List;

import net.sf.xanswer.indexing.IndexException;

import org.xiss.map.IBTree;
import org.xiss.query.IQuery;
import org.xiss.query.IResult;

/**
 * Interface for the XISS index data structure. Supports indexing of only a
 * single document. For description of the structure see <a
 * href="http://www.vldb.org/conf/2001/P361.pdf"/>
 */
public interface IXissIndex {
	/**
	 * Get a list of entries of the element index by the element name ID
	 */
	public List<ElementIndexEntry> getElementIndexEntriesByNid(int nid)
			throws IndexException;

	/**
	 * Get the whole element index as a B-tree
	 */
	public IBTree<Integer, List<ElementIndexEntry>> getElementIndex()
			throws IndexException;

	/**
	 * Get a list of entries of the attribute index by the attribute name ID
	 */
	public List<ElementIndexEntry> getAttributeIndexEntriesByNid(int nid)
			throws IndexException;

	/**
	 * Get the whole attribute index as a B-tree
	 */
	public IBTree<Integer, List<ElementIndexEntry>> getAttributeIndex()
			throws IndexException;

	/**
	 * Get the whole structure index as a random-access list of structure index
	 * entries
	 */
	public List<StructureIndexEntry> getStructureIndex() throws IndexException;

	/**
	 * Get the element index entry corresponding to the root element
	 */
	public ElementIndexEntry getRootElementIndexEntry() throws IndexException;

	/**
	 * Get the total number of nodes in the document
	 */
	public int getTotalNodesCount() throws IndexException;

	/**
	 * Get name of the node with the given ID, or null if the node does not
	 * exist
	 */
	public String getNodeName(int id) throws IndexException;

	/**
	 * Get value of the node with the given ID, or null if the node does not
	 * exist
	 */
	public String getNodeValue(int id) throws IndexException;

	/**
	 * Get the whole name table as a B-tree
	 */
	public IBTree<Integer, String> getNameTable() throws IndexException;

	/**
	 * Get name ID by the name string
	 */
	public int getNidByName(String name) throws IndexException;

	/**
	 * Get name string by its key (ID)
	 */
	public String getNameByNid(int nid) throws IndexException;

	/**
	 * Get the whole value table as a B-tree
	 */
	public IBTree<Integer, String> getValueTable() throws IndexException;

	/**
	 * Get value ID by the value string
	 */
	public int getVidByValue(String value) throws IndexException;

	/**
	 * Get value string by its key (ID)
	 */
	public String getValueByVid(int vid) throws IndexException;

	/**
	 * Run a query on the index
	 */
	public IResult query(IQuery query) throws IndexException;

	/**
	 * Close the index and flush all caches
	 */
	public void close() throws IndexException;

	/**
	 * Delete the index
	 * 
	 * @throws IndexException
	 */
	public void drop() throws IndexException;
	
	public static final int NONE = -1;
}