package org.xiss.index.impl;

import java.util.Stack;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import net.sf.xanswer.indexing.IndexException;

import org.xiss.index.IXissIndex;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

public class XmlIndexBuilder {
	private XissIndexImpl myIndex;

	private static final String TEXT_ELEMENT_NAME = "text()";

	private static final int NONE = -1;

	private int myCurrentParentId;

	private int myCurrentId;

	private Stack<ElementIndexEntryImpl> myElementStack;

	private Stack<StructureIndexEntryImpl> myStructureStack;

	private Stack<StructureIndexEntryImpl> myPreviousStructureStack;

	private StringBuilder myCurrentContentBuilder;

	private String myCurrentTextContents;

	private String myIndexDirectory;

	private String myXmlLocation;

	private final Logger myLogger = Logger.getLogger("XmlIndexBuilder");

	public XmlIndexBuilder(String xmlLocation, String indexLocation) {
		myXmlLocation = xmlLocation;
		myIndexDirectory = indexLocation;

		disableLogging();
	}

	public void disableLogging() {
		myLogger.setLevel(Level.OFF);
	}

	public void enableLogging() {
		myLogger.setLevel(Level.ALL);
	}

	private ContentHandler myHandler = new DefaultHandler() {
		boolean isRoot = true;

		{
			myCurrentContentBuilder = new StringBuilder();
		}

		public void startElement(String uri, String localName, String qName,
				Attributes attributes) {

			try {
				if (localName != XmlIndexBuilder.TEXT_ELEMENT_NAME) {
					handleText();
				}
				if (!myPreviousStructureStack.empty()) {
					StructureIndexEntryImpl previousStructure = myPreviousStructureStack
							.pop();
					previousStructure.FirstSiblingId = myCurrentId;
				}

				int eid = myCurrentId++;
				int depth = myElementStack.size();

				ElementIndexEntryImpl elementEntry = new ElementIndexEntryImpl(
						eid, 0, myElementStack.size(), myCurrentParentId);
				myElementStack.push(elementEntry);

				int nid = myIndex.updateNameTable(qName);
				StructureIndexEntryImpl structureEntry = new StructureIndexEntryImpl(
						eid,
						0, // Id, size
						nid,
						XmlIndexBuilder.NONE, // Name, value
						0, XmlIndexBuilder.NONE, XmlIndexBuilder.NONE,
						XmlIndexBuilder.NONE);
				StructureIndexEntryImpl parentEntry = (myStructureStack.empty()) ? null
						: myStructureStack.peek();
				myStructureStack.push(structureEntry);

				myIndex.addElementIndexEntry(nid, elementEntry, isRoot);
				isRoot = false;
				myIndex.addStructureIndexEntry(structureEntry);

				// Set first child for parent entry if not yet set (i.e. if we
				// are
				// the first child)
				if (parentEntry != null) {
					structureEntry.ParentId = (parentEntry.Id);
					if (parentEntry.FirstChildId == XmlIndexBuilder.NONE) {
						parentEntry.FirstChildId = eid;
					}
				}
				// Insert attributes into name/value and structure index
				myCurrentParentId = eid;
				if (attributes != null) {
					structureEntry.FirstAttributeId = (updateAttributeIndex(attributes));
				}
			} catch (IndexException e) {
				throw new RuntimeException(e);
			}
		}

		public void endElement(String uri, String localName, String qName) {
			try {
				if (localName != XmlIndexBuilder.TEXT_ELEMENT_NAME) {
					handleText();
				}

				// Set correct 'size' in the top element and structure index
				// entries
				int eid = myElementStack.peek().Id;
				int size = myCurrentId - eid - 1;

				myLogger.info("Next id == " + myCurrentId + ", Last id == "
						+ eid);
				myLogger.info("Ending element " + qName + ": size = " + size);

				int nid = myIndex.updateNameTable(qName); // Will actually not
				// update but just fetch
				// existing nid

				if (localName == XmlIndexBuilder.TEXT_ELEMENT_NAME) {
					int vid = myIndex.updateValueTable(myCurrentTextContents);
					myElementStack.peek().ValueId = (vid);
					myStructureStack.peek().ValueId = (vid);
				}
				myElementStack.peek().Size = (size);
				// Re-serialize the entry with its enclosing list
				myIndex.updateElementIndexEntry(nid, myElementStack.peek());
				myStructureStack.peek().Size = (size);
				myCurrentParentId = myElementStack.peek().ParentId;
				myLogger.info("New parent == " + myCurrentParentId);

				// Quit the element
				myElementStack.pop();
				myPreviousStructureStack.push(myStructureStack.pop());

				myIndex.incrementNodeCount();
			} catch (IndexException e) {
				throw new RuntimeException(e);
			}
		}

		public void characters(char[] ch, int start, int length) {
			myCurrentContentBuilder.append(ch, start, length);
		}

		private void handleText() {
			if (myCurrentContentBuilder != null
					&& myCurrentContentBuilder.length() > 0) {
				myCurrentTextContents = myCurrentContentBuilder.toString();
				startElement("", XmlIndexBuilder.TEXT_ELEMENT_NAME,
						XmlIndexBuilder.TEXT_ELEMENT_NAME, null);
				endElement("", XmlIndexBuilder.TEXT_ELEMENT_NAME,
						XmlIndexBuilder.TEXT_ELEMENT_NAME);
				myCurrentContentBuilder = new StringBuilder();
			}
		}

		private int updateAttributeIndex(Attributes attributes) {
			int len = attributes.getLength();
			int firstAttributeId = XmlIndexBuilder.NONE;
			for (int i = 0; i < len; ++i) {
				int id = updateAttributeIndex(attributes, i);
				if (firstAttributeId == XmlIndexBuilder.NONE) {
					firstAttributeId = id;
				}
			}
			return firstAttributeId;
		}

		private int updateAttributeIndex(Attributes attributes, int index) {
			try {
				myLogger.info("Indexing attribute: "
						+ attributes.getQName(index) + " = "
						+ attributes.getValue(index));

				int vid = myIndex.updateValueTable(attributes.getValue(index));
				int nid = myIndex.updateNameTable(attributes.getQName(index));
				int aid = myCurrentId;
				ElementIndexEntryImpl entry = new ElementIndexEntryImpl(aid, 0,
						myElementStack.size(), myCurrentParentId, vid);

				myIndex.addToAttributeIndex(nid, entry);

				StructureIndexEntryImpl structureEntry = new StructureIndexEntryImpl(
						myCurrentId,
						0,
						nid,
						vid,
						myCurrentParentId,
						XmlIndexBuilder.NONE,
						(index == attributes.getLength() - 1) ? XmlIndexBuilder.NONE
								: (myCurrentId + 1), XmlIndexBuilder.NONE);
				myIndex.addStructureIndexEntry(structureEntry);

				++myCurrentId;

				myIndex.incrementNodeCount();
				return aid;
			} catch (IndexException e) {
				throw new RuntimeException(e);
			}
		}
	};

	public IXissIndex build() throws IndexException {
		myIndex = XissIndexImpl.createNew(myIndexDirectory);
		myElementStack = new Stack<ElementIndexEntryImpl>();
		myStructureStack = new Stack<StructureIndexEntryImpl>();
		myPreviousStructureStack = new Stack<StructureIndexEntryImpl>();

		myCurrentParentId = XmlIndexBuilder.NONE;
		myCurrentId = 0;

		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser parser = factory.newSAXParser();
			XMLReader reader = parser.getXMLReader();
			reader.setContentHandler(myHandler);

			reader.parse(myXmlLocation);
		} catch (Exception e) {
			e.printStackTrace();
			throw new IndexException("Unable to create index: " + e.toString());
		}

		return myIndex;
	}
}
